/*
 * Copyright (C) 2004 Nathan Lutchansky <lutchann@litech.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <errno.h>

#include <event.h>
#include <frame.h>
#include <stream.h>
#include <outputs.h>
#include <rtp.h>
#include <conf_parse.h>

/* MPEG4 Video ES format:
 * 
 * +--Included in SDP config line
 * |
 * | Visual Object Sequence
 * |   start code (0x000001B0), profile and level indication, stuff
 * |     Visual Object
 * |       start code (0x000001B5), visual object type, stuff
 * |     Video Object
 * |       start code (0x00000100-0x0000011F)
 * |     Video Object Layer
 * |       start code (0x00000120-0x0000012F), stuff to parse
 *           Group Video Object Plane*
 *             start code (0x000001B3), time codes, stuff
 *           Video Object Plane*
 *             start code (0x000001B6), vop coding type, modulo time base,
 *             vop time increment
 *     end code (0x000001B1)
 */

struct rtp_mpeg4 {
	struct stream_destination *input;
	struct rtp_stream *rtp;
	unsigned char rtp_buf[1400];
	int rtp_len; // = 0
	unsigned char config[256];
	int config_len;
	int pali;
	char sdp_lines[512];
	int stream_ready; //  = 0;
	int vop_time_increment_resolution;
	int vop_time_increment;
	int vtir_bitlen;
	int fincr;
	int fbase;
	unsigned int rtp_seqnum; // = 0,
	unsigned int rtp_timestamp; // = 0,
	unsigned int rtp_ssrc; // = 0x87654321;
};

static void send_to_rtp( struct rtp_mpeg4 *out, struct frame *f,
				unsigned char *d, int len )
{
	int i, plen;

	if( ! f )
	{
		memcpy( out->rtp_buf + out->rtp_len, d, len );
		out->rtp_len += len;
		return;
	}
	if( out->rtp_len > 0 )
	{
		plen = 1400 - out->rtp_len;
		if( plen > len ) plen = len;
		memcpy( out->rtp_buf + out->rtp_len, d, plen );
		out->rtp_len += plen;
		send_rtp_packet( out->rtp, out->rtp_buf,
					out->rtp_len, plen == len );
		out->rtp_len = 0;
		i = plen;
	} else i = 0;
	while( i < len )
	{
		plen = len - i;
		if( plen > 1400 ) plen = 1400;
		send_rtp_packet( out->rtp, d + i, plen, plen + i == len );
		i += plen;
	}
	update_rtp_timestamp( out->rtp, 90000 * out->fincr / out->fbase );
}

static void mpeg4_check_endpoints( struct rtp_stream *rtp,
					struct rtp_endpoint *ep )
{
	struct rtp_mpeg4 *out = (struct rtp_mpeg4 *)rtp->private;
	int waiting = ! out->stream_ready;

	for( ep = rtp->ep_list; ep; ep = ep->next )
		waiting = ep->running;

	set_waiting( out->input, waiting );
	if( ! waiting ) out->rtp_len = 0;
}

static int find_next_code( unsigned char *d, int len )
{
	int i;

	/* Start at 1 to ignore the code sitting at d[0] */
	for( i = 1; i < len - 2; ++i )
		if( d[i] == 0x00 && d[i+1] == 0x00 && d[i+2] == 0x01 )
			return i;
	return 0;
}

static void add_to_config( struct rtp_mpeg4 *out, unsigned char *d, int len )
{
	memcpy( out->config + out->config_len, d, len );
	out->config_len += len;
}

static void finish_config( struct rtp_mpeg4 *out )
{
	int i;
	char *o = out->sdp_lines;

	o += sprintf( o, "a=fmtp:96 profile-level-id=%d;config=", out->pali );

	for( i = 0; i < out->config_len; ++i )
		o += sprintf( o, "%02X", out->config[i] );

	sprintf( o, "\r\n" );

	out->stream_ready = 1;
	out->input->stream->get_framerate( out->input->stream,
			&out->fincr, &out->fbase );
	send_to_rtp( out, NULL, out->config, out->config_len );
	mpeg4_check_endpoints( out->rtp, NULL );
}

static const char *mpeg4_get_sdplines( struct rtp_stream *rtp )
{
	struct rtp_mpeg4 *out = (struct rtp_mpeg4 *)rtp->private;

	return out->stream_ready ? out->sdp_lines : NULL;
}

static unsigned int get_field( unsigned char *d, int bits, int *offset )
{
	unsigned int v = 0;
	int i;

	for( i = 0; i < bits; )
	{
		if( bits - i >= 8 && *offset % 8 == 0 )
		{
			v <<= 8;
			v |= d[*offset/8];
			i += 8;
			*offset += 8;
		} else
		{
			v <<= 1;
			v |= ( d[*offset/8] >> ( 7 - *offset % 8 ) ) & 1;
			++i;
			++(*offset);
		}
	}
	return v;
}

static void parse_visual_object_sequence( struct rtp_mpeg4 *out,
				struct frame *f, unsigned char *d, int len )
{
	out->config_len = 0;
	out->pali = d[4];
	add_to_config( out, d, len );
}

static void parse_visual_object( struct rtp_mpeg4 *out, struct frame *f,
					unsigned char *d, int len )
{
	int visual_object_type;
	int off = 32;

	add_to_config( out, d, len );
	if( get_field( d, 1, &off ) ) off += 7;
	visual_object_type = get_field( d, 4, &off );
	//printf( "visual_object_type = %d\n", visual_object_type );
}

static void parse_video_object( struct rtp_mpeg4 *out, struct frame *f,
					unsigned char *d, int len )
{
	// should be empty except for the start code
	add_to_config( out, d, len );
}

static void parse_video_object_layer( struct rtp_mpeg4 *out, struct frame *f,
					unsigned char *d, int len )
{
	int off = 41, i;

	if( get_field( d, 1, &off ) ) off += 7; // object layer identifier
	if( get_field( d, 4, &off ) == 15 ) // aspect ratio info
		off += 16; // extended par
	if( get_field( d, 1, &off ) ) // vol control parameters
	{
		off += 3; // chroma format, low delay
		if( get_field( d, 1, &off ) ) off += 79; // vbw parameters
	}
	off += 2; // video object layer shape
	if( ! get_field( d, 1, &off ) )
	{
		printf( "error: missing marker!\n" );
		return;
	}
	i = out->vop_time_increment_resolution = get_field( d, 16, &off );
	if( ! get_field( d, 1, &off ) )
	{
		printf( "error: missing marker!\n" );
		return;
	}
	for( out->vtir_bitlen = 0; i != 0; i >>= 1 ) ++out->vtir_bitlen;
	if( get_field( d, 1, &off ) )
	{
		out->vop_time_increment = get_field( d, out->vtir_bitlen, &off );
		if( out->vop_time_increment == 0 )
		{
			printf( "error: fixed VOP time increment is 0!\n" );
			return;
		}
		//printf( "apparent framerate: %f\n", (double)out->vop_time_increment_resolution / (double)out->vop_time_increment );
	} else out->vop_time_increment = 0;

	add_to_config( out, d, len );
}

static void parse_video_object_plane( struct rtp_mpeg4 *out, struct frame *f,
					unsigned char *d, int len )
{
	int off = 32;
	int vop_coding_type, modulo_time_base, var_time_increment;

	vop_coding_type = get_field( d, 2, &off );
	//printf( "vop coding type is %d\n", vop_coding_type );
	for( modulo_time_base = 0; off / 8 < len; ++modulo_time_base )
		if( ! get_field( d, 1, &off ) ) break;
	if( ! get_field( d, 1, &off ) )
	{
		printf( "error: missing marker!\n" );
		return;
	}
	var_time_increment = get_field( d, out->vtir_bitlen, &off );

	if( vop_coding_type == 0 )
		printf( "frame type %c, size %d, vop time increment %d, modulo time base %d\n",
			"IPBS"[vop_coding_type], len,
			var_time_increment, modulo_time_base );

	send_to_rtp( out, f, d, len );
}

static int parse_mpeg4_frame( struct rtp_mpeg4 *out, struct frame *f,
					unsigned char *d, int len )
{
	unsigned int start_code;

	start_code = ( d[0] << 24 ) | ( d[1] << 16 ) | ( d[2] << 8 ) | d[3];
	//printf( "frame start code: %08x\n", start_code );
	if( start_code == 0x000001B0 )
		parse_visual_object_sequence( out, f, d, len );
	else if( start_code == 0x000001B5 )
		parse_visual_object( out, f, d, len );
	else if( start_code >= 0x00000100 && start_code <= 0x0000011F )
		parse_video_object( out, f, d, len );
	else if( start_code >= 0x00000120 && start_code <= 0x0000012F )
		parse_video_object_layer( out, f, d, len );
	else if( start_code == 0x000001B2 )
	{
		add_to_config( out, d, len );
		finish_config( out );
	} else if( start_code == 0x000001B6 )
	{
		parse_video_object_plane( out, f, d, len );
		return 1;
	}
	return 0;
}

static void mpeg4_next_frame( struct frame *f, void *d )
{
	struct rtp_mpeg4 *out = (struct rtp_mpeg4 *)d;
	int flen, start;

	for( start = 0; start < f->length; start += flen )
	{
		flen = find_next_code( f->d + start, f->length - start );
		if( flen == 0 ) flen = f->length - start;
		parse_mpeg4_frame( out, f, f->d + start, flen );
	}
	unref_frame( f );
}

/************************ CONFIGURATION DIRECTIVES ************************/

static void *start_block(void)
{
	struct rtp_mpeg4 *out;

	out = (struct rtp_mpeg4 *)malloc( sizeof( struct rtp_mpeg4 ) );
	out->input = NULL;
	out->rtp = (struct rtp_stream *)malloc( sizeof( struct rtp_stream ) );
	out->rtp->get_sdp = mpeg4_get_sdplines;
	out->rtp->new_endpoint = mpeg4_check_endpoints;
	out->rtp->del_endpoint = mpeg4_check_endpoints;
	out->rtp->ep_list = NULL;
	out->rtp->private = out;
	out->rtp_len = 0;
	out->stream_ready = 0;
	out->rtp_seqnum = 0;
	out->rtp_timestamp = 0;
	out->rtp_ssrc = 0x87654321;

	return out;
}

static int end_block( void *d )
{
	struct rtp_mpeg4 *out = (struct rtp_mpeg4 *)d;

	if( ! out->input )
	{
		printf( "missing input stream name\n" );
		return -1;
	}
	set_waiting( out->input, 1 );

	return 0;
}

static int set_input( int num_tokens, struct token *tokens, void *d )
{
	struct rtp_mpeg4 *out = (struct rtp_mpeg4 *)d;
	int format = FORMAT_MPEG4;

	if( ! ( out->input = connect_to_stream( tokens[1].v.str,
					mpeg4_next_frame, out, &format, 1 ) ) )
	{
		printf( "unable to connect to stream \"%s\"\n",
				tokens[1].v.str );
		return -1;
	}
	return 0;
}

int new_rtsp_location( char *path, struct rtp_stream *rtp );

static int set_path( int num_tokens, struct token *tokens, void *d )
{
	struct rtp_mpeg4 *out = (struct rtp_mpeg4 *)d;

	new_rtsp_location( tokens[1].v.str, out->rtp );
}

static struct statement config_statements[] = {
	/* directive name, process function, min args, max args, arg types */
	{ "input", set_input, 1, 1, { TOKEN_STR } },
	{ "path", set_path, 1, 1, { TOKEN_STR } },

	/* empty terminator -- do not remove */
	{ NULL, NULL, 0, 0, {} }
};

int rtp_mpeg4_init(void)
{
	register_config_context( "output", "rtp-mpeg4", start_block, end_block,
					config_statements );
	return 0;
}
