/*
 * Copyright (C) 2004 Nathan Lutchansky <lutchann@litech.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <event.h>
#include <frame.h>
#include <stream.h>
#include <rtp.h>
#include <conf_parse.h>

struct session {
	struct session *next;
	struct session *prev;
	char id[32];
	int client_port;
	int server_port;
	struct rtp_endpoint *endpoint;
};

struct conn {
	struct conn *next;
	struct conn *prev;
	int fd;
	struct sockaddr_in client_addr;
	struct event *read_event;
	struct event *write_event;
	char req_buf[4096];
	int req_len;
	struct { char *name; char *value; } req_hdr[16];
	unsigned char rep_buf[65536];
	unsigned char *rep;
	int rep_len;
	int drop_after;
};

struct rtsp_location {
	struct rtsp_location *next;
	struct rtsp_location *prev;
	char path[256];
	struct rtp_stream *rtp;
};

struct http_location {
	struct http_location *next;
	struct http_location *prev;
	char path[256];
	struct stream_destination *input;
	struct frame *frame;
};

struct listener {
	int fd;
};

struct conn *conn_list = NULL;
struct session *sess_list = NULL;
struct rtsp_location *rtsp_loc_list = NULL;
struct http_location *http_loc_list = NULL;

extern struct frame *jpeg;

int new_rtsp_location( char *path, struct rtp_stream *rtp )
{
	struct rtsp_location *loc;

	loc = (struct rtsp_location *)malloc( sizeof( struct rtsp_location ) );
	loc->next = rtsp_loc_list;
	loc->prev = NULL;
	if( loc->next ) loc->next->prev = loc;
	rtsp_loc_list = loc;
	strcpy( loc->path, path );
	loc->rtp = rtp;
	return 0;
}

static void del_session( struct session *sess )
{
	if( sess->next ) sess->next->prev = sess->prev;
	if( sess->prev ) sess->prev->next = sess->next;
	else sess_list = sess->next;
	free( sess );
}

static struct session *new_session( void )
{
	struct session *sess;
	int len;

	sess = (struct session *)malloc( sizeof( struct session ) );
	sess->next = sess_list;
	sess->prev = NULL;
	if( sess->next ) sess->next->prev = sess;
	sess_list = sess;

	sess->client_port = sess->server_port = -1;
	sess->endpoint = NULL;
	for( len = 0; len < 15; ++len )
		sprintf( sess->id + len * 2, "%02X", random() & 0xff );
	sess->id[30] = 0;
	return sess;
}

static struct session *get_session( char *id )
{
	struct session *s;

	for( s = sess_list; s; s = s->next )
		if( ! strcmp( s->id, id ) ) break;
	return s;
}

static void drop_conn( struct conn *c )
{
	remove_event( c->read_event );
	remove_event( c->write_event );
	close( c->fd );
	c->fd = -1;
	if( c->next ) c->next->prev = c->prev;
	if( c->prev ) c->prev->next = c->next;
	else conn_list = c->next;
	free( c );
}

static void conn_write( struct event *e, void *d )
{
	struct conn *c = (struct conn *)d;
	int ret;

	ret = write( c->fd, c->rep, c->rep_len );
	if( ret <= 0 )
	{
		if( ret < 0 && errno == EAGAIN ) return;
		drop_conn( c );
		return;
	}
	c->rep += ret;
	c->rep_len -= ret;
	if( c->rep_len > 0 ) return;
	*strstr( c->rep_buf, "\r\n\r\n" ) = 0;
	printf( "reply:\n%s\n", c->rep_buf );
	if( c->drop_after ) drop_conn( c );
	else set_event_enabled( c->write_event, 0 );
}

static void send_reply( struct conn *c )
{
	c->rep = c->rep_buf;
	set_event_enabled( c->write_event, 1 );
}

static char *get_header( struct conn *c, char *name )
{
	int i;
	static char *empty = "";

	for( i = 0; c->req_hdr[i].name; ++i )
		if( ! strcasecmp( c->req_hdr[i].name, name ) )
			return c->req_hdr[i].value;
	return empty;
}

static void handle_OPTIONS( struct conn *c )
{
	c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nPublic: DESCRIBE, SETUP, TEARDOWN, PLAY\r\n\r\n", get_header( c, "CSeq" ) );
	send_reply( c );
}

static void handle_DESCRIBE( struct conn *c )
{
	const char *fmtp;
	char sdp[8192];
	int sdp_len;

	fmtp = rtsp_loc_list->rtp->get_sdp( rtsp_loc_list->rtp );
	if( ! fmtp )
	{
		c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 404 Not Found\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) );
		send_reply( c );
		return;
	}
	sdp_len = sprintf( sdp, "v=0\r\no=- 1 1 IN IP4 127.0.0.1\r\ns=Test\r\na=type:broadcast\r\na=control:rtsp://192.168.16.21:7080/test/\r\nt=0 0\r\nc=IN IP4 0.0.0.0\r\nm=video 0 RTP/AVP 96\r\na=rtpmap:96 MP4V-ES/90000\r\n%sa=control:track1\r\n\r\n", fmtp );
	c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nContent-Type: application/sdp\r\nContent-Length: %d\r\n\r\n%s", get_header( c, "CSeq" ), sdp_len, sdp );
	send_reply( c );
}

static void handle_SETUP( struct conn *c )
{
	char *p;
	int cport;
	struct session *s;

	if( ! ( p = strstr( get_header( c, "Transport" ), "client_port" ) ) ||
		( *(p + 11) != '=' ) || ! ( cport = atoi( p + 12 ) ) )
	{
		c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 461 Unsupported Transport\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) );
		send_reply( c );
		return;
	}
	printf( "client port is %d\n", cport );
	s = new_session();
	s->client_port = cport;
	s->endpoint = new_rtp_udp_endpoint( rtsp_loc_list->rtp,
			c->client_addr.sin_addr, cport, &s->server_port );
	c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %s\r\nTransport: RTP/AVP;unicast;client_port=%d-%d;server_port=%d-%d\r\n\r\n", get_header( c, "CSeq" ), s->id, s->client_port, s->client_port + 1, s->server_port, s->server_port + 1 );
	send_reply( c );
}

static void handle_PLAY( struct conn *c )
{
	struct session *s;

	if( ! ( s = get_session( get_header( c, "Session" ) ) ) )
	{
		c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 454 Session Not Found\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) );
		send_reply( c );
		return;
	}
	c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %s\r\n\r\n", get_header( c, "CSeq" ), s->id );
	send_reply( c );
	s->endpoint->running = 1;
	s->endpoint->rtp_stream->new_endpoint( s->endpoint->rtp_stream, s->endpoint );
}

static void handle_TEARDOWN( struct conn *c )
{
	struct session *s;

	if( ! ( s = get_session( get_header( c, "Session" ) ) ) )
	{
		c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 454 Session Not Found\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) );
		send_reply( c );
		return;
	}
	s->endpoint->running = 0;
	s->endpoint->rtp_stream->del_endpoint( s->endpoint->rtp_stream, s->endpoint );
	del_rtp_endpoint( s->endpoint );
	c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 200 OK\r\nCSeq: %s\r\nSession: %s\r\n\r\n", get_header( c, "CSeq" ), s->id );
	send_reply( c );
	del_session( s );
}

static void handle_GET( struct conn *c )
{
	struct http_location *loc = http_loc_list;

	if( ! loc || ! loc->frame )
	{
		c->rep_len = sprintf( c->rep_buf, "HTTP/1.0 404 Not Found\r\nConnection: close\r\nContent-Type: text/html\r\n\r\n<html><body><p>Image not available</p></body></html>\n" );
		send_reply( c );
		c->drop_after = 1;
		return;
	}
	c->rep_len = sprintf( c->rep_buf, "HTTP/1.0 200 OK\r\nConnection: close\r\nContent-Type: image/jpeg\r\nContent-Length: %d\r\n\r\n", loc->frame->length );
	/* No bounds checking here -- that will be fixed when we do the
	 * improved TCP connection handling */
	memcpy( c->rep_buf + c->rep_len, loc->frame->d, loc->frame->length );
	c->rep_len += loc->frame->length;
	send_reply( c );
	c->drop_after = 1;
}

static void handle_unknown( struct conn *c )
{
	c->rep_len = sprintf( c->rep_buf, "RTSP/1.0 501 Not Implemented\r\nCSeq: %s\r\n\r\n", get_header( c, "CSeq" ) );
	send_reply( c );
}

static int handle_request( struct conn *c )
{
	char *t;
	int i = 0;

	if( c->req_buf[0] == 0 ) return -1;

	printf( "Request:\n%s\n", c->req_buf );

	/* Terminate the first line of the request */
	for( t = c->req_buf; *t != '\r' && *t != '\n'; ++t )
		if( *t == 0 ) return -1;
	*(t++) = 0;
	if( *t == '\n' ) *(t++) = 0;
	/* Find the name and content of each header line */
	for( i = 0; i < 15; ++i )
	{
		/* We should be sitting on the header name */
		c->req_hdr[i].name = t;
		/* Terminate where the colon is */
		for( ; *t != ':'; ++t )
			if( *t == 0 || *t == '\r' || *t == '\n' ) return -1;
		*(t++) = 0;
		/* Skip any spaces after the colon */
		for( ; *t == ' '; ++t )
			if( *t == 0 ) return -1;
		/* Should be the start of the header content */
		c->req_hdr[i].value = t;
		for( ; *t != '\r' && *t != '\n'; ++t )
			if( *t == 0 ) return -1;
		/* Terminate the content string */
		*(t++) = 0;
		if( *t == '\n' ) *(t++) = 0;
		/* Are we at the end of the header? */
		if( *t == 0 || *t == '\r' || *t == '\n' ) break;
	}
	if( i == 15 ) return -1; /* Too many headers */
	c->req_hdr[i + 1].name = NULL;

#if 0
	for( i = 0; i < 16 && c->req_hdr[i].name; ++i )
	{
		printf( "Header name = %s\n", c->req_hdr[i].name );
		printf( "Header value = %s\n", c->req_hdr[i].value );
	}
#endif

	if( ! strncasecmp( c->req_buf, "OPTIONS ", 8 ) )
		handle_OPTIONS( c );
	else if( ! strncasecmp( c->req_buf, "DESCRIBE ", 9 ) )
		handle_DESCRIBE( c );
	else if( ! strncasecmp( c->req_buf, "SETUP ", 6 ) )
		handle_SETUP( c );
	else if( ! strncasecmp( c->req_buf, "PLAY ", 5 ) )
		handle_PLAY( c );
	else if( ! strncasecmp( c->req_buf, "TEARDOWN ", 9 ) )
		handle_TEARDOWN( c );
	else if( ! strncasecmp( c->req_buf, "GET ", 4 ) )
		handle_GET( c );
	else handle_unknown( c );
	return 0;
}

static void do_read( struct event *e, void *d )
{
	struct conn *c = (struct conn *)d;
	char *t;
	int ret;

	ret = read( c->fd, c->req_buf + c->req_len,
				sizeof( c->req_buf ) - c->req_len - 1 );
	if( ret <= 0 )
	{
		if( ret < 0 ) perror( "read" );
		else fprintf( stderr, "connection closed\n" );
		drop_conn( c );
		return;
	}
	c->req_len += ret;
	c->req_buf[c->req_len] = 0;
	if( ( t = strstr( c->req_buf, "\r\n\r\n" ) ) )
	{
		t[2] = 0;
		if( handle_request( c ) < 0 )
		{
			fprintf( stderr, "malformed request from client\n" );
			drop_conn( c );
			return;
		}
		c->req_len = 0;
	}
	if( c->req_len == sizeof( c->req_buf ) - 1 )
	{
		fprintf( stderr, "malformed request from client\n" );
		drop_conn( c );
	}
}

static void do_accept( struct event *e, void *d )
{
	struct listener *listener = (struct listener *)d;
	int fd;
	socklen_t addrlen;
	struct sockaddr_in addr;
	struct conn *c;

	addrlen = sizeof( addr );
	if( ( fd = accept( listener->fd, (struct sockaddr *)&addr, &addrlen ) ) < 0 )
	{
		perror( "accept" );
		exit( 1 );
	}
	printf( "accepted connection\n" );

	fcntl( fd, F_SETFL, O_NONBLOCK );

	c = (struct conn *)malloc( sizeof( struct conn ) );
	c->next = conn_list;
	if( c->next ) c->next->prev = c;
	c->prev = NULL;
	c->fd = fd;
	c->client_addr = addr;
	c->req_len = 0;
	c->read_event = add_fd_event( fd, 0, 0, do_read, c );
	c->write_event = add_fd_event( fd, 1, 0, conn_write, c );
	set_event_enabled( c->write_event, 0 );
	c->rep_len = 0;
	c->drop_after = 0;
	conn_list = c;
}

int rtsp_listen( int port )
{
	struct sockaddr_in addr;
	struct listener *listener;
	int opt, fd;

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = 0;
	addr.sin_port = htons( port );

	if( ( fd = socket( PF_INET, SOCK_STREAM, 0 ) ) < 0 )
	{
		perror( "socket" );
		return -1;
	}
	opt = 1;
	if( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof( opt ) ) < 0 )
		perror( "setsockopt" );
	if( bind( fd, (struct sockaddr *)&addr, sizeof( addr ) ) < 0 )
	{
		perror( "bind" );
		close( fd );
		return -1;
	}
	if( listen( fd, 5 ) < 0 )
	{
		perror( "listen" );
		close( fd );
		return -1;
	}

	listener = (struct listener *)malloc( sizeof( struct listener ) );
	listener->fd = fd;

	add_fd_event( fd, 0, 0, do_accept, listener );

	printf( "listening on port %d\n", port );

	return 0;
}

/********************* GLOBAL CONFIGURATION DIRECTIVES ********************/

int config_port( int num_tokens, struct token *tokens, void *d )
{
	int port;

	port = tokens[1].v.num;

	if( port <= 0 || port > 65535 )
	{
		printf( "port %d is out of range!\n", port );
		return -1;
	}

	return rtsp_listen( port );
}

/************************ CONFIGURATION DIRECTIVES ************************/

static void jpeg_next_frame( struct frame *f, void *d )
{
	struct http_location *loc = (struct http_location *)d;

	if( loc->frame ) unref_frame( loc->frame );
	loc->frame = f;
}

static void *start_block(void)
{
	struct http_location *loc;

	loc = (struct http_location *)malloc( sizeof( struct http_location ) );
	loc->next = http_loc_list;
	loc->prev = NULL;
	if( loc->next ) loc->next->prev = loc;
	http_loc_list = loc;
	loc->input = NULL;
	loc->path[0] = 0;
	loc->frame = NULL;

	return loc;
}

static int end_block( void *d )
{
	struct http_location *loc = (struct http_location *)d;

	if( ! loc->input )
	{
		printf( "http output: missing input stream name\n" );
		return -1;
	}
	if( ! *loc->path )
	{
		printf( "http output: missing URL path\n" );
		return -1;
	}
	set_waiting( loc->input, 1 );

	return 0;
}

static int set_input( int num_tokens, struct token *tokens, void *d )
{
	struct http_location *loc = (struct http_location *)d;
	int format = FORMAT_JPEG;

	if( ! ( loc->input = connect_to_stream( tokens[1].v.str,
					jpeg_next_frame, loc, &format, 1 ) ) )
	{
		printf( "unable to connect to stream \"%s\"\n",
				tokens[1].v.str );
		return -1;
	}
	return 0;
}

static int set_path( int num_tokens, struct token *tokens, void *d )
{
	struct http_location *loc = (struct http_location *)d;

	strcpy( loc->path, tokens[1].v.str );
}

static struct statement config_statements[] = {
	/* directive name, process function, min args, max args, arg types */
	{ "input", set_input, 1, 1, { TOKEN_STR } },
	{ "path", set_path, 1, 1, { TOKEN_STR } },

	/* empty terminator -- do not remove */
	{ NULL, NULL, 0, 0, {} }
};

int http_init(void)
{
	register_config_context( "output", "http", start_block, end_block,
					config_statements );
	return 0;
}
